diff -u -r --new-file nginx-1.2.2-org/auto/sources nginx-1.2.2/auto/sources
--- nginx-1.2.2-org/auto/sources	2012-07-02 12:41:13.000000000 -0400
+++ nginx-1.2.2/auto/sources	2012-08-21 12:23:45.000000000 -0400
@@ -291,6 +291,7 @@
            src/http/ngx_http_script.h \
            src/http/ngx_http_upstream.h \
            src/http/ngx_http_upstream_round_robin.h \
+           src/http/ngx_http_upstream_random.h \
            src/http/ngx_http_busy_lock.h"
 
 HTTP_SRCS="src/http/ngx_http.c \
@@ -307,6 +308,7 @@
            src/http/ngx_http_script.c \
            src/http/ngx_http_upstream.c \
            src/http/ngx_http_upstream_round_robin.c \
+           src/http/ngx_http_upstream_random.c \
            src/http/ngx_http_parse_time.c \
            src/http/modules/ngx_http_static_module.c \
            src/http/modules/ngx_http_index_module.c \
diff -u -r --new-file nginx-1.2.2-org/src/http/ngx_http.h nginx-1.2.2/src/http/ngx_http.h
--- nginx-1.2.2-org/src/http/ngx_http.h	2012-01-18 10:07:43.000000000 -0500
+++ nginx-1.2.2/src/http/ngx_http.h	2012-08-21 12:23:45.000000000 -0400
@@ -29,6 +29,7 @@
 #include <ngx_http_request.h>
 #include <ngx_http_upstream.h>
 #include <ngx_http_upstream_round_robin.h>
+#include <ngx_http_upstream_random.h>
 #include <ngx_http_config.h>
 #include <ngx_http_busy_lock.h>
 #include <ngx_http_script.h>
diff -u -r --new-file nginx-1.2.2-org/src/http/ngx_http_upstream.c nginx-1.2.2/src/http/ngx_http_upstream.c
--- nginx-1.2.2-org/src/http/ngx_http_upstream.c	2012-07-02 11:43:50.000000000 -0400
+++ nginx-1.2.2/src/http/ngx_http_upstream.c	2012-08-21 12:23:45.000000000 -0400
@@ -895,7 +895,7 @@
     }
 #endif
 
-    if (ngx_http_upstream_create_round_robin_peer(r, ur) != NGX_OK) {
+    if (ngx_http_upstream_create_random_peer(r, ur) != NGX_OK) {
         ngx_http_upstream_finalize_request(r, u,
                                            NGX_HTTP_INTERNAL_SERVER_ERROR);
         return;
diff -u -r --new-file nginx-1.2.2-org/src/http/ngx_http_upstream_random.c nginx-1.2.2/src/http/ngx_http_upstream_random.c
--- nginx-1.2.2-org/src/http/ngx_http_upstream_random.c	1969-12-31 19:00:00.000000000 -0500
+++ nginx-1.2.2/src/http/ngx_http_upstream_random.c	2012-08-21 12:24:26.000000000 -0400
@@ -0,0 +1,541 @@
+
+/*
+ * Copyright (C) Igor Sysoev
+ * Copyright (C) Nginx, Inc.
+ */
+
+
+#include <ngx_config.h>
+#include <ngx_core.h>
+#include <ngx_http.h>
+
+
+static ngx_http_upstream_random_peer_t *ngx_http_upstream_get_peer(
+    ngx_http_upstream_random_peer_data_t *rp);
+
+#if (NGX_HTTP_SSL)
+
+static ngx_int_t ngx_http_upstream_empty_set_session(ngx_peer_connection_t *pc,
+    void *data);
+static void ngx_http_upstream_empty_save_session(ngx_peer_connection_t *pc,
+    void *data);
+
+#endif
+
+
+ngx_int_t
+ngx_http_upstream_init_random(ngx_conf_t *cf,
+    ngx_http_upstream_srv_conf_t *us)
+{
+    ngx_url_t                      u;
+    ngx_uint_t                     i, n;
+    ngx_http_upstream_random_peers_t  *peers;
+
+    us->peer.init = ngx_http_upstream_init_random_peer;
+
+    /* an upstream implicitly defined by proxy_pass, etc. */
+
+    if (us->port == 0 && us->default_port == 0) {
+        ngx_log_error(NGX_LOG_EMERG, cf->log, 0,
+                      "no port in upstream \"%V\" in %s:%ui",
+                      &us->host, us->file_name, us->line);
+        return NGX_ERROR;
+    }
+
+    ngx_memzero(&u, sizeof(ngx_url_t));
+
+    u.host = us->host;
+    u.port = (in_port_t) (us->port ? us->port : us->default_port);
+
+    if (ngx_inet_resolve_host(cf->pool, &u) != NGX_OK) {
+        if (u.err) {
+            ngx_log_error(NGX_LOG_EMERG, cf->log, 0,
+                          "%s in upstream \"%V\" in %s:%ui",
+                          u.err, &us->host, us->file_name, us->line);
+        }
+
+        return NGX_ERROR;
+    }
+
+    n = u.naddrs;
+
+    peers = ngx_pcalloc(cf->pool, sizeof(ngx_http_upstream_random_peers_t)
+                              + sizeof(ngx_http_upstream_random_peer_t) * (n - 1));
+    if (peers == NULL) {
+        return NGX_ERROR;
+    }
+
+    peers->single = (n == 1);
+    peers->number = n;
+    peers->name = &us->host;
+
+    for (i = 0; i < u.naddrs; i++) {
+        peers->peer[i].sockaddr = u.addrs[i].sockaddr;
+        peers->peer[i].socklen = u.addrs[i].socklen;
+        peers->peer[i].name = u.addrs[i].name;
+        peers->peer[i].max_fails = 1;
+        peers->peer[i].fail_timeout = 10;
+    }
+
+    us->peer.data = peers;
+
+    /* implicitly defined upstream has no backup servers */
+
+    return NGX_OK;
+}
+
+
+ngx_int_t
+ngx_http_upstream_init_random_peer(ngx_http_request_t *r,
+    ngx_http_upstream_srv_conf_t *us)
+{
+    ngx_uint_t                            n;
+    ngx_http_upstream_random_peer_data_t  *rp;
+
+    rp = r->upstream->peer.data;
+
+    if (rp == NULL) {
+        rp = ngx_palloc(r->pool, sizeof(ngx_http_upstream_random_peer_data_t));
+        if (rp == NULL) {
+            return NGX_ERROR;
+        }
+
+        r->upstream->peer.data = rp;
+    }
+
+    rp->peers = us->peer.data;
+    rp->current = 0;
+
+    n = rp->peers->number;
+
+    if (rp->peers->next && rp->peers->next->number > n) {
+        n = rp->peers->next->number;
+    }
+
+    if (n <= 8 * sizeof(uintptr_t)) {
+        rp->tried = &rp->data;
+        rp->data = 0;
+
+    } else {
+        n = (n + (8 * sizeof(uintptr_t) - 1)) / (8 * sizeof(uintptr_t));
+
+        rp->tried = ngx_pcalloc(r->pool, n * sizeof(uintptr_t));
+        if (rp->tried == NULL) {
+            return NGX_ERROR;
+        }
+    }
+
+    r->upstream->peer.get = ngx_http_upstream_get_random_peer;
+    r->upstream->peer.free = ngx_http_upstream_free_random_peer;
+    r->upstream->peer.tries = rp->peers->number;
+#if (NGX_HTTP_SSL)
+    r->upstream->peer.set_session =
+                               ngx_http_upstream_set_random_peer_session;
+    r->upstream->peer.save_session =
+                               ngx_http_upstream_save_random_peer_session;
+#endif
+
+    return NGX_OK;
+}
+
+
+ngx_int_t
+ngx_http_upstream_create_random_peer(ngx_http_request_t *r,
+    ngx_http_upstream_resolved_t *ur)
+{
+    u_char                            *p;
+    size_t                             len;
+    ngx_uint_t                         i, n;
+    struct sockaddr_in                *sin;
+    ngx_http_upstream_random_peers_t      *peers;
+    ngx_http_upstream_random_peer_data_t  *rp;
+
+    rp = r->upstream->peer.data;
+
+    if (rp == NULL) {
+        rp = ngx_palloc(r->pool, sizeof(ngx_http_upstream_random_peer_data_t));
+        if (rp == NULL) {
+            return NGX_ERROR;
+        }
+
+        r->upstream->peer.data = rp;
+    }
+
+    peers = ngx_pcalloc(r->pool, sizeof(ngx_http_upstream_random_peers_t)
+                     + sizeof(ngx_http_upstream_random_peer_t) * (ur->naddrs - 1));
+    if (peers == NULL) {
+        return NGX_ERROR;
+    }
+
+    peers->single = (ur->naddrs == 1);
+    peers->number = ur->naddrs;
+    peers->name = &ur->host;
+
+    if (ur->sockaddr) {
+        peers->peer[0].sockaddr = ur->sockaddr;
+        peers->peer[0].socklen = ur->socklen;
+        peers->peer[0].name = ur->host;
+        peers->peer[0].max_fails = 1;
+        peers->peer[0].fail_timeout = 10;
+
+    } else {
+
+        for (i = 0; i < ur->naddrs; i++) {
+
+            len = NGX_INET_ADDRSTRLEN + sizeof(":65536") - 1;
+
+            p = ngx_pnalloc(r->pool, len);
+            if (p == NULL) {
+                return NGX_ERROR;
+            }
+
+            len = ngx_inet_ntop(AF_INET, &ur->addrs[i], p, NGX_INET_ADDRSTRLEN);
+            len = ngx_sprintf(&p[len], ":%d", ur->port) - p;
+
+            sin = ngx_pcalloc(r->pool, sizeof(struct sockaddr_in));
+            if (sin == NULL) {
+                return NGX_ERROR;
+            }
+
+            sin->sin_family = AF_INET;
+            sin->sin_port = htons(ur->port);
+            sin->sin_addr.s_addr = ur->addrs[i];
+
+            peers->peer[i].sockaddr = (struct sockaddr *) sin;
+            peers->peer[i].socklen = sizeof(struct sockaddr_in);
+            peers->peer[i].name.len = len;
+            peers->peer[i].name.data = p;
+            peers->peer[i].max_fails = 1;
+            peers->peer[i].fail_timeout = 10;
+        }
+    }
+
+    rp->peers = peers;
+    rp->current = 0;
+
+    if (rp->peers->number <= 8 * sizeof(uintptr_t)) {
+        rp->tried = &rp->data;
+        rp->data = 0;
+
+    } else {
+        n = (rp->peers->number + (8 * sizeof(uintptr_t) - 1))
+                / (8 * sizeof(uintptr_t));
+
+        rp->tried = ngx_pcalloc(r->pool, n * sizeof(uintptr_t));
+        if (rp->tried == NULL) {
+            return NGX_ERROR;
+        }
+    }
+
+    r->upstream->peer.get = ngx_http_upstream_get_random_peer;
+    r->upstream->peer.free = ngx_http_upstream_free_random_peer;
+    r->upstream->peer.tries = rp->peers->number;
+#if (NGX_HTTP_SSL)
+    r->upstream->peer.set_session = ngx_http_upstream_empty_set_session;
+    r->upstream->peer.save_session = ngx_http_upstream_empty_save_session;
+#endif
+
+    return NGX_OK;
+}
+
+
+ngx_int_t
+ngx_http_upstream_get_random_peer(ngx_peer_connection_t *pc, void *data)
+{
+    ngx_http_upstream_random_peer_data_t  *rp = data;
+
+    ngx_int_t                      rc;
+    ngx_uint_t                     i, n;
+    ngx_connection_t              *c;
+    ngx_http_upstream_random_peer_t   *peer;
+    ngx_http_upstream_random_peers_t  *peers;
+
+    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, pc->log, 0,
+                   "get random peer, try: %ui", pc->tries);
+
+    /* ngx_lock_mutex(rp->peers->mutex); */
+
+    if (rp->peers->last_cached) {
+
+        /* cached connection */
+
+        c = rp->peers->cached[rp->peers->last_cached];
+        rp->peers->last_cached--;
+
+        /* ngx_unlock_mutex(ppr->peers->mutex); */
+
+#if (NGX_THREADS)
+        c->read->lock = c->read->own_lock;
+        c->write->lock = c->write->own_lock;
+#endif
+
+        pc->connection = c;
+        pc->cached = 1;
+
+        return NGX_OK;
+    }
+
+    pc->cached = 0;
+    pc->connection = NULL;
+
+    if (rp->peers->single) {
+        peer = &rp->peers->peer[0];
+
+    } else {
+
+        /* there are several peers */
+
+        peer = ngx_http_upstream_get_peer(rp);
+
+        if (peer == NULL) {
+            goto failed;
+        }
+
+        ngx_log_debug1(NGX_LOG_DEBUG_HTTP, pc->log, 0,
+                       "get random peer, current: %ui",
+                       rp->current);
+    }
+
+    pc->sockaddr = peer->sockaddr;
+    pc->socklen = peer->socklen;
+    pc->name = &peer->name;
+
+    /* ngx_unlock_mutex(rp->peers->mutex); */
+
+    if (pc->tries == 1 && rp->peers->next) {
+        pc->tries += rp->peers->next->number;
+    }
+
+    return NGX_OK;
+
+failed:
+
+    peers = rp->peers;
+
+    if (peers->next) {
+
+        /* ngx_unlock_mutex(peers->mutex); */
+
+        ngx_log_debug0(NGX_LOG_DEBUG_HTTP, pc->log, 0, "backup servers");
+
+        rp->peers = peers->next;
+        pc->tries = rp->peers->number;
+
+        n = rp->peers->number / (8 * sizeof(uintptr_t)) + 1;
+        for (i = 0; i < n; i++) {
+             rp->tried[i] = 0;
+        }
+
+        rc = ngx_http_upstream_get_random_peer(pc, rp);
+
+        if (rc != NGX_BUSY) {
+            return rc;
+        }
+
+        /* ngx_lock_mutex(peers->mutex); */
+    }
+
+    /* all peers failed, mark them as live for quick recovery */
+
+    for (i = 0; i < peers->number; i++) {
+        peers->peer[i].fails = 0;
+    }
+
+    /* ngx_unlock_mutex(peers->mutex); */
+
+    pc->name = peers->name;
+
+    return NGX_BUSY;
+}
+
+
+static ngx_http_upstream_random_peer_t *
+ngx_http_upstream_get_peer(ngx_http_upstream_random_peer_data_t *rp)
+{
+    time_t                        now;
+    ngx_uint_t                    i, j;
+    ngx_http_upstream_random_peer_t  *peer, *best;
+
+    now = ngx_time();
+
+    best = NULL;
+
+    j = (int) ( (float)(rp->peers->number) * (rand() / (RAND_MAX + 1.0)));
+
+    for (i = j; i < rp->peers->number; i++) {
+        peer = &rp->peers->peer[i];
+        if (peer->down) {
+            continue;
+        }
+        best = peer;
+        break;
+    }
+
+    if (best == NULL) {
+        for (i = 0; i < j; i++) {
+            peer = &rp->peers->peer[i];
+            if (peer->down) {
+                continue;
+            }
+            best = peer;
+            break;
+        }
+    }
+
+    if (best == NULL) {
+        return NULL;
+    }
+
+    i = best - &rp->peers->peer[0];
+
+    rp->current = i;
+
+    best->checked = now;
+
+    return best;
+}
+
+
+void
+ngx_http_upstream_free_random_peer(ngx_peer_connection_t *pc, void *data,
+    ngx_uint_t state)
+{
+    ngx_http_upstream_random_peer_data_t  *rp = data;
+
+    time_t                           now;
+    ngx_http_upstream_random_peer_t  *peer;
+
+    ngx_log_debug2(NGX_LOG_DEBUG_HTTP, pc->log, 0,
+                   "free random peer %ui %ui", pc->tries, state);
+
+    if (state == 0 && pc->tries == 0) {
+        return;
+    }
+
+    /* TODO: NGX_PEER_KEEPALIVE */
+
+    if (rp->peers->single) {
+        pc->tries = 0;
+        return;
+    }
+
+    peer = &rp->peers->peer[rp->current];
+
+    if (state & NGX_PEER_FAILED) {
+        now = ngx_time();
+
+        /* ngx_lock_mutex(rp->peers->mutex); */
+
+        peer->fails++;
+        peer->accessed = now;
+        peer->checked = now;
+
+        ngx_log_debug1(NGX_LOG_DEBUG_HTTP, pc->log, 0,
+                       "free random peer failed: %ui",
+                       rp->current);
+
+        /* ngx_unlock_mutex(rp->peers->mutex); */
+
+    } else {
+
+        /* mark peer live if check passed */
+
+        if (peer->accessed < peer->checked) {
+            peer->fails = 0;
+        }
+    }
+
+    if (pc->tries) {
+        pc->tries--;
+    }
+
+    /* ngx_unlock_mutex(rp->peers->mutex); */
+}
+
+
+#if (NGX_HTTP_SSL)
+
+ngx_int_t
+ngx_http_upstream_set_random_peer_session(ngx_peer_connection_t *pc,
+    void *data)
+{
+    ngx_http_upstream_random_peer_data_t  *rp = data;
+
+    ngx_int_t                     rc;
+    ngx_ssl_session_t            *ssl_session;
+    ngx_http_upstream_random_peer_t  *peer;
+
+    peer = &rp->peers->peer[rp->current];
+
+    /* TODO: threads only mutex */
+    /* ngx_lock_mutex(rp->peers->mutex); */
+
+    ssl_session = peer->ssl_session;
+
+    rc = ngx_ssl_set_session(pc->connection, ssl_session);
+
+    ngx_log_debug2(NGX_LOG_DEBUG_HTTP, pc->log, 0,
+                   "set session: %p:%d",
+                   ssl_session, ssl_session ? ssl_session->references : 0);
+
+    /* ngx_unlock_mutex(rp->peers->mutex); */
+
+    return rc;
+}
+
+
+void
+ngx_http_upstream_save_random_peer_session(ngx_peer_connection_t *pc,
+    void *data)
+{
+    ngx_http_upstream_random_peer_data_t  *rp = data;
+
+    ngx_ssl_session_t                *old_ssl_session, *ssl_session;
+    ngx_http_upstream_random_peer_t  *peer;
+
+    ssl_session = ngx_ssl_get_session(pc->connection);
+
+    if (ssl_session == NULL) {
+        return;
+    }
+
+    ngx_log_debug2(NGX_LOG_DEBUG_HTTP, pc->log, 0,
+                   "save session: %p:%d", ssl_session, ssl_session->references);
+
+    peer = &rp->peers->peer[rp->current];
+
+    /* TODO: threads only mutex */
+    /* ngx_lock_mutex(rp->peers->mutex); */
+
+    old_ssl_session = peer->ssl_session;
+    peer->ssl_session = ssl_session;
+
+    /* ngx_unlock_mutex(rp->peers->mutex); */
+
+    if (old_ssl_session) {
+
+        ngx_log_debug2(NGX_LOG_DEBUG_HTTP, pc->log, 0,
+                       "old session: %p:%d",
+                       old_ssl_session, old_ssl_session->references);
+
+        /* TODO: may block */
+
+        ngx_ssl_free_session(old_ssl_session);
+    }
+}
+
+
+static ngx_int_t
+ngx_http_upstream_empty_set_session(ngx_peer_connection_t *pc, void *data)
+{
+    return NGX_OK;
+}
+
+
+static void
+ngx_http_upstream_empty_save_session(ngx_peer_connection_t *pc, void *data)
+{
+    return;
+}
+
+#endif
diff -u -r --new-file nginx-1.2.2-org/src/http/ngx_http_upstream_random.h nginx-1.2.2/src/http/ngx_http_upstream_random.h
--- nginx-1.2.2-org/src/http/ngx_http_upstream_random.h	1969-12-31 19:00:00.000000000 -0500
+++ nginx-1.2.2/src/http/ngx_http_upstream_random.h	2012-08-21 12:24:33.000000000 -0400
@@ -0,0 +1,84 @@
+
+/*
+ * Copyright (C) Igor Sysoev
+ * Copyright (C) Nginx, Inc.
+ */
+
+
+#ifndef _NGX_HTTP_UPSTREAM_RANDOM_H_INCLUDED_
+#define _NGX_HTTP_UPSTREAM_RANDOM_H_INCLUDED_
+
+
+#include <ngx_config.h>
+#include <ngx_core.h>
+#include <ngx_http.h>
+
+
+typedef struct {
+    struct sockaddr                *sockaddr;
+    socklen_t                       socklen;
+    ngx_str_t                       name;
+
+    ngx_uint_t                      fails;
+    time_t                          accessed;
+    time_t                          checked;
+
+    ngx_uint_t                      max_fails;
+    time_t                          fail_timeout;
+
+    ngx_uint_t                      down;          /* unsigned  down:1; */
+
+#if (NGX_HTTP_SSL)
+    ngx_ssl_session_t              *ssl_session;   /* local to a process */
+#endif
+} ngx_http_upstream_random_peer_t;
+
+
+typedef struct ngx_http_upstream_random_peers_s  ngx_http_upstream_random_peers_t;
+
+struct ngx_http_upstream_random_peers_s {
+    ngx_uint_t                      number;
+    ngx_uint_t                      last_cached;
+
+ /* ngx_mutex_t                    *mutex; */
+    ngx_connection_t              **cached;
+
+    unsigned                        single:1;
+
+    ngx_str_t                      *name;
+
+    ngx_http_upstream_random_peers_t   *next;
+
+    ngx_http_upstream_random_peer_t     peer[1];
+};
+
+
+typedef struct {
+    ngx_http_upstream_random_peers_t   *peers;
+    ngx_uint_t                      current;
+    uintptr_t                      *tried;
+    uintptr_t                       data;
+} ngx_http_upstream_random_peer_data_t;
+
+
+ngx_int_t ngx_http_upstream_init_random(ngx_conf_t *cf,
+    ngx_http_upstream_srv_conf_t *us);
+ngx_int_t ngx_http_upstream_init_random_peer(ngx_http_request_t *r,
+    ngx_http_upstream_srv_conf_t *us);
+ngx_int_t ngx_http_upstream_create_random_peer(ngx_http_request_t *r,
+    ngx_http_upstream_resolved_t *ur);
+ngx_int_t ngx_http_upstream_get_random_peer(ngx_peer_connection_t *pc,
+    void *data);
+void ngx_http_upstream_free_random_peer(ngx_peer_connection_t *pc,
+    void *data, ngx_uint_t state);
+
+#if (NGX_HTTP_SSL)
+ngx_int_t
+    ngx_http_upstream_set_random_peer_session(ngx_peer_connection_t *pc,
+    void *data);
+void ngx_http_upstream_save_random_peer_session(ngx_peer_connection_t *pc,
+    void *data);
+#endif
+
+
+#endif /* _NGX_HTTP_UPSTREAM_RANDOM_H_INCLUDED_ */
